@@ -1,3 +1,4 @@
+from autogen.oai.client import OpenAIWrapper
 from autogen.oai.completion import Completion, ChatCompletion
 from autogen.oai.openai_utils import (
     get_config_list,
@@ -5,9 +6,11 @@ from autogen.oai.openai_utils import (
     config_list_openai_aoai,
     config_list_from_models,
     config_list_from_json,
+    config_list_from_dotenv,
 )
 
 __all__ = [
+    "OpenAIWrapper",
     "Completion",
     "ChatCompletion",
     "get_config_list",
@@ -15,4 +18,5 @@ __all__ = [
     "config_list_openai_aoai",
     "config_list_from_models",
     "config_list_from_json",
+    "config_list_from_dotenv",
 ]
